//========================================================================
//
// XPDFViewer.cc
//
// Copyright 2002-2003 Glyph & Cog, LLC
//
//========================================================================

#include <aconf.h>

#ifdef USE_GCC_PRAGMAS
#pragma implementation
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gmem.h"
#include "gfile.h"
#include "GString.h"
#include "GList.h"
#include "Error.h"
#include "GlobalParams.h"
#include "PDFDoc.h"
#include "Link.h"
#include "ErrorCodes.h"
#include "Outline.h"
#include "UnicodeMap.h"
#include "XPDFApp.h"
#include "XPDFViewer.h"
#include "PSOutputDev.h"
#include "config.h"

#include <btron/hmi.h>
#include <btron/libapp.h>
#include <tstring.h>

#include <btron/proctask.h>

#include "SplashBitmap.h"
#include "CoreOutputDev.h"
#include "SplashOutputDev.h"
#include "databox.h"
#include "pict.h"
#include "fusen.h"
#include "Vobj.h"

extern XpdfFusen fsn;

struct ZoomMenuInfo {
  char *label;
  double zoom;
};

#define maxZoomIdx   0
#define defZoomIdx   3
#define minZoomIdx   7
#define zoomPageIdx  8
#define zoomWidthIdx 9

XPDFViewer::XPDFViewer(XPDFApp *appA, LINK *link,
		       int pageA, GString *destName,
		       GString *ownerPassword, GString *userPassword) 
{
  LinkDest *dest = 0;
  int pg = pageA;
  double z;
  app = appA;
  wid = subwid = -1;
  sw_fst_id = sw_bwd_id = sw_fwd_id = sw_lst_id = -1;
  nb_page_id = -1;
  nb_page_val = pg;

  core = NULL;
  tc_title = 0;
  vobjs = 0;
  vsel = 0;

  // do Motif-specific initialization and create the window;
  // this also creates the core object
  initWindow();

  if (link) {
    if (loadFile(link, ownerPassword, userPassword)) {
      getPageAndDest(pageA, destName, &pg, &dest);
      if (pg > 0) {
	core->resizeToPage(pg);
      }
    } else {
      cchg_par(sw_fst_id, P_DISABLE);
      cchg_par(sw_bwd_id, P_DISABLE);
      cchg_par(sw_fwd_id, P_DISABLE);
      cchg_par(sw_lst_id, P_DISABLE);
      return;
    }
  }

  // display the first page
  z = core->getZoom();
  if (dest) {
    displayDest(dest, z, core->getRotate(), gTrue);
    delete dest;
  } else {
    displayPage(pg, z, core->getRotate(), gTrue, gTrue);
  }

  setButtonAttr();
}

XPDFViewer::~XPDFViewer() 
{
  delete vsel;
  delete vobjs;
  delete core;
  closeWindow();
}

void XPDFViewer::open(LINK *link, int pageA, GString *destName) 
{
  LinkDest *dest;
  int pg;
  double z;

  if (!core->getDoc() /*|| fileName->cmp(core->getDoc()->getFileName())*/) {
    if (!loadFile(link, NULL, NULL)) {
      return;
    }
  }
  getPageAndDest(pageA, destName, &pg, &dest);
//  z = app->getFullScreen() ? zoomPage : core->getZoom();
  z = core->getZoom();
  if (dest) {
    displayDest(dest, z, core->getRotate(), gTrue);
    delete dest;
  } else {
    displayPage(pg, z, core->getRotate(), gTrue, gTrue);
  }
}

void XPDFViewer::clear() 
{
  char *title;

  core->clear();

  // set up title, number-of-pages display
  title = app->getTitle() ? app->getTitle()->getCString()
                          : (char *)xpdfAppName;
  setTitle(title);
}

// added for BTRON
void XPDFViewer::setTitle(const char *s)
{
  int len = strlen(s);

  delete [] tc_title;
  tc_title = 0;

  tc_title = new TC[len+1];
  eucstotcs(tc_title, s);
  wset_tit(wid, -1, tc_title, 0);
}

//------------------------------------------------------------------------
// load / display
//------------------------------------------------------------------------

GBool XPDFViewer::loadFile(LINK *link, GString *ownerPassword,
			   GString *userPassword) 
{
  GBool ret = (core->loadFile(link, ownerPassword, userPassword) == errNone);

  vobjs = new VobjSet(core->getDoc()->getNumPages());
  vobjs->load(wid, link);

  return ret;
}

void XPDFViewer::displayPage(int pageA, double zoomA, int rotateA,
			     GBool scrollToTop, GBool addToHist) 
{
  core->displayPage(pageA, zoomA, rotateA, scrollToTop, addToHist);
}

void XPDFViewer::displayDest(LinkDest *dest, double zoomA, int rotateA,
			     GBool addToHist) 
{
  core->displayDest(dest, zoomA, rotateA, addToHist);
}

void XPDFViewer::getPageAndDest(int pageA, GString *destName,
				int *pageOut, LinkDest **destOut) 
{
  Ref pageRef;

  // find the page number for a named destination
  *pageOut = pageA;
  *destOut = NULL;
  if (destName && (*destOut = core->getDoc()->findDest(destName))) {
    if ((*destOut)->isPageRef()) {
      pageRef = (*destOut)->getPageRef();
      *pageOut = core->getDoc()->findPage(pageRef.num, pageRef.gen);
    } else {
      *pageOut = (*destOut)->getPageNum();
    }
  }

  if (*pageOut <= 0) {
    *pageOut = 1;
  }
  if (*pageOut > core->getDoc()->getNumPages()) {
    *pageOut = core->getDoc()->getNumPages();
  }
}

//------------------------------------------------------------------------
// GUI code: main window
//------------------------------------------------------------------------

void XPDFViewer::initWindow() 
{
  char *title;
  RECT wrect, subwrect;
  UW attr = WA_NORMAL | WA_WORK | (fsn.full ? WA_FULL : 0);
  W bbar_pid, rbar_pid;

  wrect = fsn.wr;
  subwrect = fsn.subwr;
  subwrect.c.right = subwrect.c.left + CTL_WINDOW_WIDTH;
  subwrect.c.bottom = subwrect.c.top + 32;

  // top-level window & frame
  title = app->getTitle() ? app->getTitle()->getCString()
                          : (char *)xpdfAppName;
  delete [] tc_title;
  tc_title = new TC[strlen(title) + 1];
  eucstotcs(tc_title, title);
  
  wget_inf(WI_PANELBACK, &background, sizeof(PAT));

  wid = wopn_wnd(attr, 0 /* no parent */, 
                 &wrect, NULL, 10, tc_title, &background, NULL);
  subwid = wopn_wnd(WA_SUBW | WA_TITL | WA_HHDL | WA_WORK, wid, 
                    &subwrect, NULL, 0, tc_title, &background, NULL);

  setrect(sw_fst_r, CTL_WINDOW_GAP, CTL_WINDOW_GAP, 
          CTL_WINDOW_GAP + CTL_SWITCH_WIDTH, CTL_SWITCH_BOTTOM);
  setrect(sw_bwd_r, sw_fst_r.c.right, CTL_WINDOW_GAP,
          sw_fst_r.c.right + CTL_SWITCH_WIDTH, CTL_SWITCH_BOTTOM);
  setrect(nb_page_r, sw_bwd_r.c.right + CTL_NBOX_GAP, CTL_WINDOW_GAP,
          sw_bwd_r.c.right + CTL_NBOX_GAP + CTL_NBOX_WIDTH, 
          CTL_SWITCH_BOTTOM);
  setrect(sw_fwd_r, nb_page_r.c.right + CTL_NBOX_GAP + CTL_MAXPAGE_WIDTH,
          CTL_WINDOW_GAP, nb_page_r.c.right + CTL_NBOX_GAP 
          + CTL_MAXPAGE_WIDTH + CTL_SWITCH_WIDTH, 
          CTL_SWITCH_BOTTOM);
  setrect(sw_lst_r, sw_fwd_r.c.right, CTL_WINDOW_GAP,
          sw_fwd_r.c.right + CTL_SWITCH_WIDTH, CTL_SWITCH_BOTTOM);

  wget_bar(wid, &rbar_pid, &bbar_pid, NULL);
  cchg_par(rbar_pid, P_DRAGBREAK|P_NORMAL|P_ENABLE);
  cchg_par(bbar_pid, P_DRAGBREAK|P_NORMAL|P_ENABLE);

  ddef_ldt(BMAP_DATA, 110, &pict_gofirst, sizeof(pictogram_dbox_entry));
  ddef_ldt(BMAP_DATA, 111, &pict_goback, sizeof(pictogram_dbox_entry));
  ddef_ldt(BMAP_DATA, 112, &pict_gonext, sizeof(pictogram_dbox_entry));
  ddef_ldt(BMAP_DATA, 113, &pict_golast, sizeof(pictogram_dbox_entry));

  sw_fst_id = ccre_pms(subwid, PM_PARTS, &sw_fst_r, 110, 0);
  sw_bwd_id = ccre_pms(subwid, PM_PARTS, &sw_bwd_r, 111, 0);
  sw_fwd_id = ccre_pms(subwid, PM_PARTS, &sw_fwd_r, 112, 0);
  sw_lst_id = ccre_pms(subwid, PM_PARTS, &sw_lst_r, 113, 0);
  nb_page_id = ccre_nbx(subwid, NB_PARTS, &nb_page_r, 0x1040, &nb_page_val, 0);

  wget_sts(wid, &rectsav, NULL);

  // core
  core = new XPDFCore(app->getPaperRGB(), app->getPaperPixel(),
                      app->getMattePixel(), app->getReverseVideo(),
                      app->getInstallCmap(), app->getRGBCubeSize());

  vsel = new VobjSelection(wget_gid(wid), core->getZoom());
}

void XPDFViewer::closeWindow() 
{
  if (wid >= 0) {
    WDSTAT stat;
    wget_sts(wid, &stat, NULL);
    if (fsn.full) {
      fsn.wr = rectsav.r;
    }
    else {
      fsn.wr = stat.r;
    }

    WDSTAT statsub;
    wget_sts(subwid, &statsub, NULL);
    fsn.subwr = statsub.r;

    // Ǥ wchg_wnd ʤΤ (oend_prc Ԥɬפ)
  }
}

void XPDFViewer::updateScreen(GID wgid)
{
  SplashBitmap *bmp = 
    (core->curTile != 0) ? core->curTile->bitmap : 0;
  
  if (bmp != 0) {
    GID bgid = bmp->getDrawEnv();
    gcop_bmp(bgid, bmp->getRect(), wgid, bmp->getRect(),
             0, G_STORE | G_CVFORM | G_CVCOLOR);
  }
}

void XPDFViewer::setButtonAttr()
{
  if (core != 0 && core->getDoc() != 0) {
    int maxp = core->getDoc()->getNumPages();
    int p = core->getPageNum();

    if (p > 1) {
      cchg_par(sw_fst_id, P_ENABLE);
      cchg_par(sw_bwd_id, P_ENABLE);
    }
    else {
      cchg_par(sw_fst_id, P_DISABLE);
      cchg_par(sw_bwd_id, P_DISABLE);
    }

    if (p < maxp) {
      cchg_par(sw_fwd_id, P_ENABLE);
      cchg_par(sw_lst_id, P_ENABLE);
    }
    else {
      cchg_par(sw_fwd_id, P_DISABLE);
      cchg_par(sw_lst_id, P_DISABLE);
    }
  }
  else {
    cchg_par(sw_fst_id, P_DISABLE);
    cchg_par(sw_bwd_id, P_DISABLE);
    cchg_par(sw_fwd_id, P_DISABLE);
    cchg_par(sw_lst_id, P_DISABLE);
  }
}

void XPDFViewer::setPageDisp(int prevPage, double prevZoom)
{
  RECT rzero = {{ 0, 0, 0, 0 }};

  nb_page_val = core->getPageNum();

  if (prevPage != nb_page_val || core->getZoom() != prevZoom) {
    core->displayPage(nb_page_val, core->getZoom(), 
                      core->getRotate(), gTrue, 
                      ((core->getZoom() != prevZoom) ? gFalse : gTrue));
    if (prevPage != nb_page_val) {
      vsel->clear();
      cset_val(nb_page_id, 1, &nb_page_val);
      setButtonAttr();
    }
    vsel->changeZoom(core->getZoom());
    
    wset_wrk(getWID(), &rzero);
    app->setWorkArea(&rzero);
    app->redraw(NULL, this);
    wreq_dsp(getSubWID());
  }
}

Vobj* XPDFViewer::registerVobj(const VLINK* vlink, const VOBJSEG* vseg, 
                               const PNT& pos, int page)
{
  return vobjs->add(getWID(), vlink, vseg, pos, page, 
                    core->getPageNum(), core->getZoom());
}

void XPDFViewer::drawVobj() const
{
  vobjs->displayPage(core->getPageNum(), core->getZoom());
}

Vobj* XPDFViewer::findVobj(const PNT& p)
{
  PNT zoomp;
  zoomp.x = p.x * 100.0 / core->getZoom();
  zoomp.y = p.y * 100.0 / core->getZoom();
  return vobjs->findVobj(core->getPageNum(), zoomp);
}

Vobj* XPDFViewer::findSelectedVobj(const PNT& p)
{
  PNT zoomp;
  zoomp.x = p.x * 100.0 / core->getZoom();
  zoomp.y = p.y * 100.0 / core->getZoom();
  return vsel->findVobj(zoomp);
}

