//========================================================================
//
// xpdf.cc
//
// Copyright 1996-2003 Glyph & Cog, LLC
//
//========================================================================

#include <aconf.h>
#include "gtypes.h"
#include "GString.h"
#include "parseargs.h"
#include "gfile.h"
#include "gmem.h"
#include "GlobalParams.h"
#include "Object.h"
#include "XPDFApp.h"
#include "config.h"

#include <basic.h>
#include <tstring.h>
#include <errcode.h>
#include <btron/proctask.h>
#include <btron/vobj.h>
#include "fileio.h"
#include "fusen.h"
#include "GList.h"
#include "XPDFViewer.h"
#include "XPDFDisp.h"
#include "singletf.h"
#include "toriaezu.h"

//------------------------------------------------------------------------
// command line options
//------------------------------------------------------------------------

static GBool contView = gFalse;
static char enableT1libStr[16] = "";
static char enableFreeTypeStr[16] = "";
static char antialiasStr[16] = "";
static char psFileArg[256];
static char paperSize[15] = "";
static int paperWidth = 0;
static int paperHeight = 0;
static GBool level1 = gFalse;
static char textEncName[128] = "";
static char textEOL[16] = "";
static char ownerPassword[33] = "\001";
static char userPassword[33] = "\001";
static GBool fullScreen = gFalse;
static char remoteName[100] = "xpdf_";
static GBool doRemoteReload = gFalse;
static GBool doRemoteRaise = gFalse;
static GBool doRemoteQuit = gFalse;
static GBool printCommands = gFalse;
#ifdef DEBUG
static GBool quiet = gFalse;
#else
static GBool quiet = gTrue;
#endif
static char cfgFileName[256] = "";
static GBool printVersion = gFalse;
static GBool printHelp = gFalse;

static ArgDesc argDesc[] = {
  {"-g",          argStringDummy, NULL,           0,
   "initial window geometry"},
  {"-geometry",   argStringDummy, NULL,           0,
   "initial window geometry"},
  {"-title",      argStringDummy, NULL,           0,
   "window title"},
  {"-cmap",       argFlagDummy,   NULL,           0,
   "install a private colormap"},
  {"-rgb",        argIntDummy,    NULL,           0,
   "biggest RGB cube to allocate (default is 5)"},
  {"-rv",         argFlagDummy,   NULL,           0,
   "reverse video"},
  {"-papercolor", argStringDummy, NULL,           0,
   "color of paper background"},
  {"-z",          argStringDummy, NULL,           0,
   "initial zoom level (percent, 'page', 'width')"},
  {"-cont",       argFlag,        &contView,      0,
   "start in continuous view mode" },
#if HAVE_T1LIB_H
  {"-t1lib",      argString,      enableT1libStr, sizeof(enableT1libStr),
   "enable t1lib font rasterizer: yes, no"},
#endif
#if HAVE_FREETYPE_FREETYPE_H | HAVE_FREETYPE_H
  {"-freetype",   argString,      enableFreeTypeStr, sizeof(enableFreeTypeStr),
   "enable FreeType font rasterizer: yes, no"},
#endif
  {"-aa",         argString,      antialiasStr,   sizeof(antialiasStr),
   "enable font anti-aliasing: yes, no"},
  {"-ps",         argString,      psFileArg,      sizeof(psFileArg),
   "default PostScript file name or command"},
  {"-paper",      argString,      paperSize,      sizeof(paperSize),
   "paper size (letter, legal, A4, A3, match)"},
  {"-paperw",     argInt,         &paperWidth,    0,
   "paper width, in points"},
  {"-paperh",     argInt,         &paperHeight,   0,
   "paper height, in points"},
  {"-level1",     argFlag,        &level1,        0,
   "generate Level 1 PostScript"},
  {"-enc",    argString,   textEncName,    sizeof(textEncName),
   "output text encoding name"},
  {"-eol",    argString,   textEOL,        sizeof(textEOL),
   "output end-of-line convention (unix, dos, or mac)"},
  {"-opw",        argString,      ownerPassword,  sizeof(ownerPassword),
   "owner password (for encrypted files)"},
  {"-upw",        argString,      userPassword,   sizeof(userPassword),
   "user password (for encrypted files)"},
  {"-fullscreen", argFlag,        &fullScreen,    0,
   "run in full-screen (presentation) mode"},
  {"-remote",     argString,      remoteName + 5, sizeof(remoteName) - 5,
   "start/contact xpdf remote server with specified name"},
  {"-reload",     argFlag,        &doRemoteReload, 0,
   "reload xpdf remove server window (with -remote only)"},
  {"-raise",      argFlag,        &doRemoteRaise, 0,
   "raise xpdf remote server window (with -remote only)"},
  {"-quit",       argFlag,        &doRemoteQuit,  0,
   "kill xpdf remote server (with -remote only)"},
  {"-cmd",        argFlag,        &printCommands, 0,
   "print commands as they're executed"},
  {"-q",          argFlag,        &quiet,         0,
   "don't print any messages or errors"},
  {"-cfg",        argString,      cfgFileName,    sizeof(cfgFileName),
   "configuration file to use in place of .xpdfrc"},
  {"-v",          argFlag,        &printVersion,  0,
   "print copyright and version info"},
  {"-h",          argFlag,        &printHelp,     0,
   "print usage information"},
  {"-help",       argFlag,        &printHelp,     0,
   "print usage information"},
  {"--help",  argFlag,     &printHelp,     0,
   "print usage information"},
  {"-?",      argFlag,     &printHelp,     0,
   "print usage information"},
  {NULL}
};

//------------------------------------------------------------------------

XpdfFusen fsn = {
  sizeof(XpdfFusen) - sizeof(UH),
  0, 1, 10000, {{50, 50, 600, 500}}, {{480, 40, -1, -1}}
};


static void readConfigFile()
{
    char buf[128];

    globalParams = new GlobalParams(cfgFileName);
    if (quiet) {
        globalParams->setErrQuiet(quiet);
    }

    globalParams->setupBaseFonts(NULL);

    if (contView) {
        globalParams->setContinuousView(contView);
    }
    if (psFileArg[0]) {
        globalParams->setPSFile(psFileArg);
    }
    if (paperSize[0]) {
        if (!globalParams->setPSPaperSize(paperSize)) {
            fprintf(stderr, "Invalid paper size\n");
        }
    } else {
        if (paperWidth) {
            globalParams->setPSPaperWidth(paperWidth);
        }
        if (paperHeight) {
            globalParams->setPSPaperHeight(paperHeight);
        }
    }
    if (level1) {
        globalParams->setPSLevel(psLevel1);
    }
    if (textEncName[0]) {
        globalParams->setTextEncoding(textEncName);
    }
    if (textEOL[0]) {
        if (!globalParams->setTextEOL(textEOL)) {
            fprintf(stderr, "Bad '-eol' value on command line\n");
        }
    }
    if (enableT1libStr[0]) {
        if (!globalParams->setEnableT1lib(enableT1libStr)) {
            fprintf(stderr, "Bad '-t1lib' value on command line\n");
        }
    }
    if (enableFreeTypeStr[0]) {
        if (!globalParams->setEnableFreeType(enableFreeTypeStr)) {
            fprintf(stderr, "Bad '-freetype' value on command line\n");
        }
    }
    if (antialiasStr[0]) {
        if (!globalParams->setAntialias(antialiasStr)) {
            fprintf(stderr, "Bad '-aa' value on command line\n");
        }
    }
    if (printCommands) {
        globalParams->setPrintCommands(printCommands);
    }

    if (fsn.zoomx100 > 0) {
        sprintf(buf, "%5.2f", ((double)fsn.zoomx100) / 100.0);
        globalParams->setInitialZoom(buf);
    }
}


static W MAIN_DISPREQ(M_DISPREQ* req)
{
    XpdfFusen fusenbuf;
    XPDFApp *app;
    GString *userPasswordStr, *ownerPasswordStr;
    W fd = -1, err;

    // ¿Ȳȥޥ͡δ
    UH* ptr = reinterpret_cast<UH*>(req->info);
    int size = *ptr + sizeof(UH);
    if (size > sizeof(XpdfFusen)) {
        size = sizeof(XpdfFusen);
    }
    memcpy(&fusenbuf, ptr, size);
    fsn = fusenbuf;

    // read config file
    readConfigFile();

    // check for password string(s)
    ownerPasswordStr = 
        ownerPassword[0] != '\001' ? new GString(ownerPassword)
        : (GString *)NULL;
    userPasswordStr = 
        userPassword[0] != '\001' ? new GString(userPassword)
        : (GString *)NULL;

    opendbox(&(req->self), USER_DATA, 0x2000);

    XPDFDisp disp(req->gid, req->lnk, fsn);
    err = disp.draw(ownerPasswordStr, userPasswordStr);

    oend_req(req->vid, (err >= 0) ? 0 : 1);
    if (userPasswordStr) {
        delete userPasswordStr;
    }
    if (ownerPasswordStr) {
        delete ownerPasswordStr;
    }

    closedbox();
    TF_Singleton::finalize();

    // check for memory leaks
    Object::memCheck(stderr);
    gMemReport(stderr);
    return err;
}


static W MAIN_EXECREQ(M_EXECREQ* req)
{
    XpdfFusen fusenbuf;
    XPDFApp *app;
    GString *userPasswordStr, *ownerPasswordStr;
    int exitCode;
    int argc = 1;
    char *argv[] = {"xpdf", 0};
    W fd = -1, err;
    XPDFViewer *viewer = 0;
    W wid, subwid;

    if ((req->mode & 3) == 1) {
        // 䵸ͭǡɤ߹
        fd = opn_fil(&(req->lnk), F_READ, NULL);
        err = oget_fsn(req->vid, fd, &fusenbuf, 
                       sizeof(XpdfFusen));
        cls_fil(fd);
        fsn = fusenbuf;
    }
    else if ((req->mode & 3) == 0) {
        // ¿Ȳȥޥ͡δ
        UH* ptr = reinterpret_cast<UH*>(req->info);
        int size = *ptr + sizeof(UH);
        if (size > sizeof(XpdfFusen)) {
            size = sizeof(XpdfFusen);
        }
        memcpy(&fusenbuf, ptr, size);
        fsn = fusenbuf;
    }
    else {
        oend_prc(req->vid, NULL, 0); // exit
        return 1;
    }

    // read config file
    readConfigFile();

    // check for password string(s)
    ownerPasswordStr = 
        ownerPassword[0] != '\001' ? new GString(ownerPassword)
        : (GString *)NULL;
    userPasswordStr = 
        userPassword[0] != '\001' ? new GString(userPassword)
        : (GString *)NULL;

    opendbox(&(req->self), USER_DATA, 0x2000);

    app = new XPDFApp(&argc, argv);
    if (!app->openLink(&(req->lnk), fsn.page, 
                       ownerPasswordStr, userPasswordStr)) {
        exitCode = 1;
        goto done;
    }

    if (app->getViewers()->getLength() > 0) {
        viewer = (XPDFViewer*)app->getViewers()->get(0);
        wid = viewer->getWID(); 
        subwid = viewer->getSubWID();
        osta_prc(req->vid, wid);
    }

    try {
        app->run();
        exitCode = 0;
    }
    catch (...) {
        exitCode = 1;
    }

 done:
    W gid = oend_prc(req->vid, &fsn, 1);
    if (gid > 0) {
        XPDFDisp disp(gid, req->lnk, fsn);
        disp.draw(ownerPasswordStr, userPasswordStr);
    }

    delete app;
    delete globalParams;

    wcls_wnd(wid, CLR);
    wcls_wnd(subwid, CLR);

    fd = opn_fil(&(req->lnk), F_UPDATE, NULL);
    err = oput_fsn(req->vid, fd, &fsn);
    cls_fil(fd);

    closedbox();
    TF_Singleton::finalize();

    if (userPasswordStr) {
        delete userPasswordStr;
    }
    if (ownerPasswordStr) {
        delete ownerPasswordStr;
    }

    // check for memory leaks
    Object::memCheck(stderr);
    gMemReport(stderr);
    return exitCode;
}


extern "C" W MAIN(MESSAGE *msg)
{
    switch (msg->msg_type) {
    case DISPREQ:
        return MAIN_DISPREQ((M_DISPREQ*)msg);
    case PASTEREQ:
        oend_req(((M_PASTEREQ*)msg)->vid, 1);
        return 1;
    case FUSENREQ:
        oend_req(((M_FUSENREQ*)msg)->vid, 1);
        return 1;
    case TADREQ:
        return ER_NOSPT;
    case EXECREQ:
        return MAIN_EXECREQ((M_EXECREQ*)msg);
    default:
        // ¾ξ
        return 1;
    }
}

